@@ -1,9 +1,10 @@
 from copy import copy
+from dataclasses import dataclass
 from json import JSONDecoder
 from typing import Union, Any, List, Optional, Type, Callable, Dict
 
 import yaml
-from jsonasobj import JsonObj, as_json, ExtendedNamespace, as_dict, JsonObjTypes, JsonTypes, items
+from jsonasobj import JsonObj, as_json, as_dict, JsonObjTypes, items
 from rdflib import Graph
 from yaml.constructor import ConstructorError
 
@@ -11,6 +12,7 @@ from linkml_runtime.utils.context_utils import CONTEXTS_PARAM_TYPE, merge_contex
 
 YAMLObjTypes = Union[JsonObjTypes, "YAMLRoot"]
 
+
 class YAMLMark(yaml.error.Mark):
     def __str__(self):
         snippet = self.get_snippet()
@@ -20,16 +22,24 @@ class YAMLMark(yaml.error.Mark):
             where += ":\n"+snippet
         return where
 
-
-class YAMLRoot(ExtendedNamespace):
-
+class YAMLRoot(JsonObj):
     """
     The root object for all python YAML representations
     """
+    def __init__(self, *args, **kwargs):
+        """
+        Override dataclass initializer
+        @param args:
+        @param kwargs:
+        """
+        super().__init__(*args, **kwargs)
 
-    def __post_init__(self, **kwargs):
-        if kwargs:
+    def __post_init__(self, *args: List[str],  **kwargs):
+        if args or kwargs:
             messages: List[str] = []
+            for v in args:
+                v = repr(v)[:40].replace('\n', '\\n')
+                messages.append(f"Unknown positional argument: {v}")
             for k in kwargs.keys():
                 v = repr(kwargs[k])[:40].replace('\n', '\\n')
                 messages.append(f"{TypedNode.yaml_loc(k)} Unknown argument: {k} = {v}")
@@ -73,100 +83,149 @@ class YAMLRoot(ExtendedNamespace):
                         rval[k] = v
             return rval
         else:
-            return obj._default(obj) if hasattr(obj, '_default') and callable(obj._default) else JSONDecoder().decode(obj)
+            return obj._default(obj) if hasattr(obj, '_default') and callable(obj._default) else\
+                JSONDecoder().decode(obj)
 
-    def _normalize_inlined_slot(self, slot_name: str, slot_type: Type, key_name: Optional[str],
-                                inlined_as_list: Optional[bool], keyed: bool) -> None:
+    def _normalize_inlined_as_list(self, slot_name: str, slot_type: Type, key_name: str, keyed: bool) -> None:
+        self._normalize_inlined(slot_name, slot_type, key_name, keyed, True)
+
+    def _normalize_inlined_as_dict(self, slot_name: str, slot_type: Type, key_name: str, keyed: bool) -> None:
+        self._normalize_inlined(slot_name, slot_type, key_name, keyed, False)
+
+    def _normalize_inlined(self, slot_name: str, slot_type: Type, key_name: str, keyed: bool, is_list: bool) \
+            -> None:
         """
          __post_init__ function for a list of inlined keyed or identified classes.
+
         The input to this is either a list or dictionary of dictionaries.  In the list case, every key entry
         in the list must be unique.  In the dictionary case, the key may or may not be repeated in the dictionary
         element. The internal storage structure is a dictionary of dictionaries.
         @param slot_name: Name of the slot being normalized
         @param slot_type: Slot range type
         @param key_name: Name of the key or identifier in the range
-        @param inlined_as_list: True means represent as a list, false or None as a dictionary
         @param keyed: True means each identifier must be unique
+        @param is_list: True means inlined as list
         """
-        raw_slot: Union[list, dict] = self[slot_name]
-        if isinstance(raw_slot, JsonObj):
-            raw_slot = as_dict(raw_slot)
-        cooked_slot = list() if inlined_as_list else dict()
-        key_list = list()
+        raw_slot: Union[list, dict, JsonObj] = self[slot_name]
+        cooked_slot = list() if is_list else dict()
+        cooked_keys = set()
 
-        def cook_a_slot(entry) -> None:
-            if keyed:
-                if key in key_list:
-                    raise ValueError(f"{TypedNode.yaml_loc(key)}: duplicate key")
-                key_list.append(key)
-            if inlined_as_list:
-                cooked_slot.append(entry)
+        def order_up(key: Any, cooked_entry: YAMLRoot) -> None:
+            """ A cooked entry is ready to be added to the return slot """
+            if cooked_entry[key_name] != key:
+                raise ValueError(
+                    f"Slot: {loc(slot_name)} - attribute {loc(key_name)} " \
+                    f"value ({loc(cooked_entry[key_name])}) does not match key ({loc(key)})")
+            if keyed and key in cooked_keys:
+                raise ValueError(f"{loc(key)}: duplicate key")
+            cooked_keys.add(key)
+            if is_list:
+                cooked_slot.append(cooked_entry)
             else:
-                cooked_slot[entry[key_name]] = entry
+                cooked_slot[key] = cooked_entry
+
+        def loc(s):
+            loc_str = TypedNode.yaml_loc(s) if isinstance(s, TypedNode) else ''
+            if loc_str == ': ':
+                loc_str = ''
+            return loc_str + str(s)
+
+        def form_1(entries: Dict[Any, Optional[Union[dict, JsonObj]]]) -> None:
+            """ A dictionary of key:dict entries where key is the identifier and dict is an instance of slot_type """
+            for key, raw_obj in items(entries):
+                if raw_obj is None:
+                    raw_obj = {}
+                if key_name not in raw_obj:
+                    raw_obj = copy(raw_obj)
+                    raw_obj[key_name] = key
+                if not issubclass(type(raw_obj), slot_type):
+                    order_up(key, slot_type(**as_dict(raw_obj)))
+                else:
+                    order_up(key, raw_obj)
+
+        # TODO: Make an external function extract a root JSON list
+        if isinstance(raw_slot, JsonObj):
+            raw_slot = raw_slot._hide_list()
 
         if isinstance(raw_slot, list):
-            # A list of dictionaries
-            #   [ {key_name: v11, slot_2: v12, ..., slot_n:v1n}, {key_name: v21, slot_2: v22, ..., slot_n:v2n}, ...]
-            for raw_slot_entry in raw_slot:
-                if not isinstance(raw_slot_entry, (dict, YAMLRoot)):
-                    raise ValueError(f"Slot: {slot_name} - unrecognized element: {raw_slot_entry}")
-                if keyed and key_name in raw_slot_entry:
-                    value = raw_slot_entry if isinstance(raw_slot_entry, slot_type) else \
-                            slot_type(**raw_slot_entry) if isinstance(raw_slot_entry, dict) else\
-                            slot_type(**raw_slot_entry.__dict__)
-                    key = getattr(value, key_name)
-                    cook_a_slot(value)
-                else:
-                    for k, v in raw_slot_entry.items():
-                        key = k
-                        cook_a_slot(slot_type(k, v))
-        elif isinstance(raw_slot, dict):
-            # A dictionary
-            # One of:
-            #    {key_1: {[key_name: v11], slot_2: v12, ... slot_n: v1n}, key_2: {...}}   or
-            #    {v11: v12, v21: v22, ...}
-            for key, value in raw_slot.items():
-                if not isinstance(value, (dict, YAMLRoot)):
-                    cook_a_slot(slot_type(key, value))
-                elif isinstance(value, YAMLRoot):
-                    vk = getattr(value, key_name, None)
-                    if vk is None or vk == key:
-                        setattr(value, key_name, key)
-                        cook_a_slot(value)
+            # We have a list of entries
+            for list_entry in raw_slot:
+                if isinstance(list_entry, (dict, JsonObj)):
+                    # list_entry is either a key:dict, key_name:value or **kwargs
+                    if len(list_entry) == 1:
+                        # key:dict or key_name:key
+                        for lek, lev in items(list_entry):
+                            if lek == key_name and not isinstance(lev, (list, dict, JsonObj)):
+                                # key_name:value
+                                order_up(list_entry[lek], slot_type(list_entry))
+                                break   # Not strictly necessary, but
+                            elif not isinstance(lev, (list, dict, JsonObj)):
+                                # key: value --> slot_type(key, value)
+                                order_up(lek, slot_type(lek, lev))
+                            else:
+                                form_1(list_entry)
                     else:
-                        raise ValueError(f"Slot: {slot_name} - value ({vk}) does not match key ({key})")
+                        # **kwargs
+                        cooked_obj = slot_type(**as_dict(list_entry))
+                        order_up(cooked_obj[key_name], cooked_obj)
+                elif isinstance(list_entry, list):
+                    # *args
+                    cooked_obj = slot_type(*list_entry)
+                    order_up(cooked_obj[key_name], cooked_obj)
                 else:
-                    # Inject a key if not there otherwise make sure it matches
-                    vk = value.get(key_name, None)
-                    if vk is None:
-                        value_copy = value.copy()
-                        value_copy[key_name] = key
-                        cook_a_slot(slot_type(**value_copy))
-                    elif vk != key:
-                        raise ValueError(f"Slot: {slot_name} - value ({vk}) does not match key ({key})")
-                    else:
-                        cook_a_slot(slot_type(**value))
+                    # lone key [key1: , key2: ... }
+                    order_up(list_entry, slot_type(**{key_name: list_entry}))
         else:
-            raise ValueError(f"Slot: {slot_name} must be a dictionary or a list")
+            # We have a dictionary
+            if key_name in raw_slot and not isinstance(raw_slot[key_name], (list, dict, JsonObj)):
+                # Vanilla dictionary - {key: v11, s12: v12, ...}
+                order_up(raw_slot[key_name], slot_type(**as_dict(raw_slot)))
+            else:
+                # We have either {key1: {obj1}, key2: {obj2}...} or {key1:, key2:, ...}
+                for k, v in items(raw_slot):
+                    if v is None:
+                        v = dict()
+                    if isinstance(v, (dict, JsonObj)):
+                        form_1({k: v})
+                    elif not isinstance(v, list):
+                        order_up(k, slot_type(*[k, v]))
+                    else:
+                        raise ValueError(f"Unrecognized entry: {loc(k)}: {str(v)}")
         self[slot_name] = cooked_slot
 
-    @staticmethod
-    def _as_list(value: List[JsonObjTypes]) -> List[JsonTypes]:
-        """ Return a json array as a list
-
-        :param value: array
-        :return: array with JsonObj instances removed
+    def _normalize_inlined_slot(self, slot_name: str, slot_type: Type, key_name: Optional[str],
+                                inlined_as_list: Optional[bool], keyed: bool) -> None:
         """
-        return [e._as_dict if isinstance(e, YAMLRoot) else e for e in value]
+        A deprecated entry point to slot normalization. Used for models generated prior to the linkml-runtime split.
 
-    @property
-    def _as_dict(self) -> Dict[str, JsonTypes]:
-        """ Convert a JsonObj into a straight dictionary
+        This code is invoked via one of the following patterns:
+            if self.<slot> is None:
+                self.<slot> = []
+            if not isinstance(self.<slot>, list):
+                self.extensions = [self.<slot>]
+            self._normalize_inlined_slot(slot_name="<slot>", slot_type=<type>, key_name="<key>", inlined_as_list=True, keyed=...)
+        or
+            if self.<slot> is None:
+                self.<slot> = []
+            if not isinstance(self.<slot>, (list, dict)):
+                self.local_names = [self.<slot>]
+            self._normalize_inlined_slot(slot_name="<slot>", slot_type=<type>, key_name="<key>", inlined_as_list=None, keyed=...)
+
+        The above pattern broke when the new jsonasobj was introduced, which is why we have the normalization above.  The code
+        below reverse engineers the above and invokes the new form
 
-        :return: dictionary that cooresponds to the json object
         """
-        return {k: v._as_dict if isinstance(v, YAMLRoot) else self._as_list(v) if isinstance(v, list) else v
-                for k, v in items(self)}
+        if inlined_as_list:
+            list_slot = self[slot_name]
+            if len(list_slot) == 1 and not isinstance(list_slot[0], list):
+                self[slot_name] = list_slot[0]
+            self._normalize_inlined_as_list(slot_name, slot_type, key_name, keyed)
+        else:
+            dict_slot = self[slot_name]
+            if isinstance(dict_slot, list):
+                if len(dict_slot) == 1 and not isinstance(dict_slot[0], (list, dict)):
+                    self[slot_name] = dict_slot[0]
+            self._normalize_inlined_as_dict(slot_name, slot_type, key_name, keyed)
 
 
 def root_representer(dumper: yaml.Dumper, data: YAMLRoot):
